import os
from glob import glob
from jinja2 import Environment, BaseLoader
from .splinter import Splinter
from .script import Script
from .renderer import Renderer
from random import randint
from sys import maxsize

MAX_SIZE = int(1e12)

class OptiBot:
    def __init__(self):
        self.splinter_registry : dict[str, Splinter] = dict()
        self.script_registry : dict[str, Script] = dict()
        self.splinters_params = 0
        self.splinter_blocks = 0

    def preload_splinters_from_path(self, base_path):
        splinters_paths = glob(base_path)

        for splinter_path in splinters_paths:
            with open(splinter_path, "r") as fp:
                splinter_parts = fp.read().split("---")
                
                if len(splinter_parts) < 2:
                    raise SyntaxError(f"Missing meta splinter in {splinter_path}")

                meta, source = splinter_parts[0], splinter_parts[1]
                splinter = self._build_splinter(meta, source)
                
                if splinter.id is None:
                    raise ValueError(f"Splinter id not defined in '{splinter_path}'")

                if splinter.id in self.splinter_registry:
                    raise ValueError(f"Splinter id {splinter.id} already exists. Duplicate found in {splinter_path}")

                self.splinter_registry[splinter.id] = splinter
    
    
    def _build_splinter(self, splinter_meta, splinter_source):
        splinter = Splinter()
        splinter.source = splinter_source
        env = Environment(loader=BaseLoader)
                
        env.globals['optiid'] = splinter.optiid_handler
        env.globals['optidesc'] = splinter.optidesc_handler
        env.globals['optitask'] = splinter.optitask_handler
        env.globals['optiimport'] = splinter.optiimport_handler
        env.globals['optivar'] = splinter.optivar_handler
        env.globals['optiparam'] = splinter.optiparam_handler

        rtemplate = env.from_string(splinter_meta)
        rtemplate.render()
        
        return splinter

    
    def preload_templates_from_path(self, base_path):
        scripts_path = glob(base_path)

        for script_path in scripts_path:
            with open(script_path, "r") as fp:
                script_source = fp.read()
                
                script_id = os.path.basename(script_path)
                script = self._build_script(script_source)
                
                if script_id in self.script_registry:
                    raise ValueError(f"Splinter id {script_id} already exists. Duplicate found in {script_path}")

                self.script_registry[script_id] = script
    

    def _build_script(self, script_source):
        script = Script(script_source)

        env = Environment(loader=BaseLoader)
                
        env.globals['splinter'] = script.splinter_handler
        env.globals['imports'] = script.imports_handler
        
        rtemplate = env.from_string(script_source)   
        rtemplate.render()

        return script


    def compile(self):
        """
        Compiles the script registry by finding the most suitable splinters 
        for each script reference by signature and task.
        Raises ValueError if no suitable splinter is found
        """

        # Initialize two variables to keep track of total splinter params and blocks
        self.splinters_params = 0
        self.splinter_blocks = 0

        # Iterate over each script in the script registry
        for script_id, script in self.script_registry.items():
            
            # Iterate over each reference in the script
            for ref in script.refs.values():
                # Initialize some variables for each reference
                is_workable = False
                max_param_size = 0
                tasks = set(ref.tasks.split("."))
                signature = set(ref.vars.keys())

                # Iterate over each splinter in the splinter registry
                for splinter_id, splinter in self.splinter_registry.items():
                    # Check if the tasks of the reference are a subset of the tasks of the splinter
                    if not tasks.issubset(splinter.tasks):
                        continue

                    # Check if the signature of the reference matches the signature of the splinter
                    if signature != splinter.signature:
                        continue

                    # If we've found a suitable splinter, update some variables for the reference
                    is_workable = True
                    max_param_size = max(max_param_size, splinter.param_size)
                    ref.splinters.add(splinter_id)

                # Set the maximum parameter size for the reference
                ref.max_param_size = max_param_size

                # If we couldn't find a suitable splinter for the reference, raise a ValueError
                if not is_workable:
                    raise ValueError(f"No suitable splinters for script '{script_id}'. task: '{ref.tasks}'. signature: '{signature}'")

                # Update the total splinter params for each script
                script.splinters_params += max_param_size
            
            self.splinters_params = max(self.splinters_params, script.splinters_params)
            self.splinter_blocks = max(self.splinter_blocks, script.splinter_blocks)


    def generate_script(self):
        for x in range(self.splinter_blocks):
            yield randint(0, MAX_SIZE)
        for x in range(self.splinters_params):
            yield randint(-MAX_SIZE, MAX_SIZE)
    
    
    def render(self, script_id, sequence):

        renderer = Renderer(self.script_registry[script_id], sequence, self.splinter_registry)
        return renderer.render()


from typing import Optional
from jinja2 import Environment, BaseLoader
from .script import Script
from .splinter import Splinter
from random import randint


IMPORT_PLACEHOLDER = "##__IMPORTS__##"

class Renderer():

    def __init__(self, script : Script, script_sequence: list[int], splinters: dict[str, Splinter]):
        self.script = script
        self.script_sequence = script_sequence
        self.splinters = splinters
        self.block_at = -1
        self.param_at = -1
        self.env : Optional[Environment] = None
        self.imports_needed = set()


    def render(self):
        self.env = Environment(loader=BaseLoader)

        self.env.globals['splinter'] = self.splinter_handler
        self.env.globals['imports'] = self.imports_handler
        
        rendered_source = self.env.from_string(self.script.source).render()

        rendered_source = rendered_source.replace(IMPORT_PLACEHOLDER, "\n".join(self.imports_needed))

        return rendered_source


    def imports_handler(self):
        return IMPORT_PLACEHOLDER

    def splinter_handler(self, ref_name, splinter_tasks, **kwargs):
        ref = self.script.refs.get(ref_name)

        self.block_at += 1
        splinter_gene = self.script_sequence[self.block_at]
        splinter_gene = int(splinter_gene % len(ref.splinters))

        splinter_to_render = list(ref.splinters)[splinter_gene]

        splinter = self.splinters.get(splinter_to_render)

        render_params = dict()
        for var_name in splinter.vars.keys():
            render_params[var_name] = ref.vars[var_name]

        for param_name in splinter.params.keys():
            self.param_at += 1
            param_gene = self.script_sequence[self.script.splinter_blocks + self.param_at]

            render_params[param_name] = self.gene_to_value(param_gene, param_name, splinter)

        rendererd_splinter = self.env.from_string(splinter.source) \
            .render(**render_params)
        
        self.imports_needed = self.imports_needed.union(splinter.imports)

        return rendererd_splinter
    
    def gene_to_value(self, value, param_name, splinter):
        param_type = splinter.params[param_name][1]
        if param_type == "int":
            return self._gene_to_value_int(value, param_name, splinter)
        if param_type == "float":
            return self._gene_to_value_float(value, param_name, splinter)
        if param_type == "str":
            return self._gene_to_value_str(value, param_name, splinter)
        
        return value
    
    def _gene_to_value_int(self, value, param_name, splinter):
        from .optibot import MAX_SIZE

        opts = splinter.params[param_name][-1]

        lowest_possible = -MAX_SIZE
        highest_possible = MAX_SIZE
        max_range = highest_possible - lowest_possible
        original_delta = (value - lowest_possible) / max_range

        if "min" in opts:
            lowest_possible = max(lowest_possible, opts["min"])

        if "max" in opts:
            highest_possible = min(highest_possible, opts["max"])

        scaled_range = highest_possible - lowest_possible
        
        scaled_value = lowest_possible + (scaled_range * original_delta)

        return int(scaled_value)

    def _gene_to_value_float(self, value, param_name, splinter):
        from .optibot import MAX_SIZE

        opts = splinter.params[param_name][-1]

        lowest_possible = -MAX_SIZE
        highest_possible = MAX_SIZE
        max_range = highest_possible - lowest_possible
        original_delta = (value - lowest_possible) / max_range

        if "ratio" in opts and opts["ratio"]:
            return original_delta


        if "min" in opts:
            lowest_possible = max(lowest_possible, opts["min"])

        if "max" in opts:
            highest_possible = min(highest_possible, opts["max"])

        scaled_range = highest_possible - lowest_possible
        
        scaled_value = lowest_possible + (scaled_range * original_delta)

        return scaled_value

    def _gene_to_value_str(self, value, param_name, splinter):
        from .optibot import MAX_SIZE

        opts = splinter.params[param_name][-1]
        
        value = value % len(opts["choices"])

        return f'\"{opts["choices"][value]}\"'



from functools import cached_property

class Splinter:
    
    def __init__(self):
        self.id = None
        self.vars = dict()
        self.params = dict()
        self.imports = set()
        self.tasks = set()
        self.source = None
        self.description = ""

        self._param_size = None
        self._signature = None
    
    @cached_property
    def signature(self):
        return set(self.vars.keys())
    
    @cached_property
    def param_size(self):
        return len(self.params)

    def optiid_handler(self, splinter_id):
        if self.id is not None:
            raise ValueError(f"Splinter {self.id} already has id defined.")

        self.id = splinter_id
    
    def optidesc_handler(self, desciption):
        self.description = desciption
    
    def optitask_handler(self, task: str):
        self.tasks.add(task)

    def optiimport_handler(self, import_str: str):
        self.imports.add(import_str)

    def optivar_handler(self, var_name: str, required : bool =False):
        if var_name in self.vars:
            raise ValueError(f"Splinter '{self.id}' already has var '{var_name}' defined ")

        self.vars[var_name] = (var_name, required)

    def optiparam_handler(self, param_name: str, param_type: str, param_values: list = None, **kwargs):
        if param_name in self.params:
            raise ValueError(f"Splinter '{self.id}' already has param '{param_name}' defined ")

        self.params[param_name] = (param_name, param_type, param_values, kwargs)
    
    
    




from __future__ import annotations
from functools import cached_property

class Script():
    
    def __init__(self, source):
        self.refs : dict[str, SplinterRef] = dict()
        self.source : str = source
        self.splinters_params = 0

    
    
    @cached_property
    def splinter_blocks(self):
        return len(self.refs.keys())
    
    def imports_handler(self):
        return ""
    
    def splinter_handler(self, ref_name, splinter_tasks, **kwargs):
        
        if ref_name in self.refs:
            raise SyntaxError(f"Script alread has a ref with name {ref_name}")
        
        self.refs[ref_name] = SplinterRef(ref_name, splinter_tasks, dict(**kwargs))


class SplinterRef():
    
    def __init__(self, name, tasks, vars):
        self.name = name
        self.tasks = tasks
        self.vars = vars
        self.splinters = set()
        self.max_param_size = 0

